/*
 *  TalKT
 *  Copyright (C) 1999 Jochen Sarrazin, DG6VJ. All rights reserved.
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "config.h"
#include "ax25k.h"

#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <errno.h>

#include <qsocketnotifier.h>

#include "toolbox.h"
#include <time.h>
#include <linux/rose.h>   // Wenn dieses #include in den ax25-libs drin ist,
                          // kann es hier 'raus.
// #include <procutils.h>
#ifdef HAVE_AX25_AXUTILS_H
   #include <ax25/axconfig.h>
   #include <ax25/axutils.h>
#else
   #include <axconfig.h>
   #include <axutils.h>
#endif

#include "main.h"
#include "ax25k.moc"

#include <qsocketnotifier.h>


extern TopLevel *toplevel;


int ax25init_done = 0;


/*
 *	ax25 -> ascii conversion
 */
char *ax25asc(ax25_address *a)
{
	static char buf[11];
	char c, *s;
	int n;

	for (n = 0, s = buf; n < 6; n++) {
		c = (a->ax25_call[n] >> 1) & 0x7F;

		if (c != ' ') *s++ = c;
	}
	
	*s++ = '-';

	if ((n = ((a->ax25_call[6] >> 1) & 0x0F)) > 9) {
		*s++ = '1';
		n -= 10;
	}
	
	*s++ = n + '0';
	*s++ = '\0';

	return buf;
}






AX25::AX25() : QObject( 0 )
{
   fd = -1;
   timer = NULL;
   if (!ax25init_done)
     if (ax25_config_load_ports() == 0)
       {
	 printf("TalKT: No AX.25 port data avialable\n");       
	 return;
       }
   ax25init_done=1;

}


AX25::~AX25()
{
   if (fd != -1)
   {
      delete timer;
      ::close(fd);
      delete sockrx;
   }
}


void AX25::connectionGot( int _fd )
{
   fd = _fd;

   timer = new QTimer( this );
   connect(timer, SIGNAL(timeout()), this, SLOT(poll()));
   timer->start(200, true);

   sockrx = new QSocketNotifier( fd, QSocketNotifier::Read, this );
   connect( sockrx, SIGNAL(activated(int)), this, SLOT(rxFrame(int)) );
}


void AX25::connectIt( char *call, char *path, char *port, char *mycall )
{
   int error;
   int paclen;


   if (ax25_config_get_addr(port) == NULL)
   {
      printf("TalKT: Invalid port setting for port %s.\n", port);
      return;
   }

   paclen = ax25_config_get_paclen(port);

   if ((fd = makeConnect( call, path, port, &error, mycall )) == -1)
   {
      switch (error)
      {
         case 1: // invalid parameter
                 printf("invalid parameter\n");
                 break;
         case 2: // cannot open socket
                 printf("cannot open socket\n");
                 break;
         case 3: // cannot bind socket
                 printf("cannot bind socket\n");
                 break;
         case 4: // cannot connect twice
                 printf("cannot connect twice\n");
                 break;
         case 0xFF: // unknown port name
                 printf("unknown port name\n");
                 break;
      }
      fd = -1;
      return;
   }


   timer = new QTimer( this );
   connect(timer, SIGNAL(timeout()), this, SLOT(poll()));
   timer->start(200, true);

   sockrx = new QSocketNotifier( fd, QSocketNotifier::Read, this );
   connect( sockrx, SIGNAL(activated(int)), this, SLOT(rxFrame(int)) );
}


int AX25::makeConnect( char *call, char *digis, char *port, int *error, char *mycall )
{
  struct full_sockaddr_ax25 ax25;
  int addrlen;
  struct full_sockaddr_ax25 ax25_d;
  int addrlen_d;
  int fd,i;
  char tmp[500];

  if (digis == NULL)
     strcpy(tmp, call);
  else if (digis[0] == '\0')
          strcpy(tmp, call);
       else
          sprintf(tmp,"%s %s",call,digis);

  addrlen_d = convert_call(tmp,&ax25_d);

  if (addrlen_d == -1)
  {
     *error = 1;
     return -1;
  }


  fd = socket(AF_AX25,SOCK_SEQPACKET,0);
  if (fd < 0)
  {
     *error = 2;
     return -1;
  }
  
  ax25.fsa_ax25.sax25_family = AF_AX25;
  ax25.fsa_ax25.sax25_ndigis = 1;
  convert_call_entry(mycall,
                     ax25.fsa_ax25.sax25_call.ax25_call);
  convert_call_entry(ax25_config_get_addr(port),
                     ax25.fsa_digipeater[0].ax25_call);
  addrlen = sizeof(struct full_sockaddr_ax25);
  
  
  if (bind(fd,(struct sockaddr *)&ax25,addrlen) == -1)
  {
    ::close(fd);
    *error = 3;
    return -1;
  }
  
  i = TRUE;
  ioctl(fd,FIONBIO,&i);

  if (::connect(fd,(struct sockaddr *)&ax25_d,addrlen_d)) {
    if (errno != EINPROGRESS) {
      ::close(fd);
      *error = 4;
      return -1;
    }
  }
  
  return(fd);
}


void AX25::poll()
{
   int axstate;
   int tries,/*unack,*/squeue;
   ax25_info_struct axinfo;
   int err;


   if ((err = ioctl(fd, SIOCAX25GETINFO, &axinfo)) != -1)
   {
      tries = axinfo.n2count;
/*      if (axinfo.vs < axinfo.va)
         unack = 8 + p->vs - p->va;
      else
         unack = p->vs - p->va;*/
      squeue = axinfo.snd_q / 256;
      if ((squeue == 0) && (axinfo.snd_q > 0)) squeue = 1;

      switch (axinfo.state)
      {
         case 0:
         case 1:
           axstate = axinfo.state;
           break;
         case 2:
           axstate = 3;
           break;
         case 3:
           axstate = 4;
           break;
        case 4:
           axstate = 6;
           break;
         default:
           axstate = 0;
           break;
       }

          // Aenderungen an die Channel-Klasse zurueckgeben
//          tmp->channel->setUnack(unack);
/*          tmp->channel->setSQueue(squeue);
          tmp->channel->setTries(tries);
          if (axstate == 0)
             disc.append(tmp->channel);
          else*/
             toplevel->setStatus(axstate);
   }
/*   else
   {
//         disc.append(tmp->channel);
   }*/

   timer->start(200, true);
}


//   void Channel::rxFrame(int socket)
//
// Diese Funktion wird aufgerufen, wenn von unserer Verbindung Daten
// gekommen sind.
void AX25::rxFrame(int socket)
{
   char tmp[1001];
   int len;
   
   if ((len = read(socket, tmp, 1000))<0) {
     disc();
     return;
   }
   if (len == 0) return;
   toplevel->workWithFrame( tmp, len );
}


//   void AX25::disc()
//
// Disconected die Verbindung
void AX25::disc()
{
   ::close(fd);
   fd = -1;
   delete sockrx;
   sockrx = NULL;

   if (timer != NULL)
   {
      delete timer;
      timer = NULL;
   }
}



///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////


ListenAX25::ListenAX25( const char *_call, char *_port ) : QObject( 0 )
{
   int i;
   struct full_sockaddr_ax25 ax25;
   int addrlen;
   char *addr;
   if (!ax25init_done)
     if (ax25_config_load_ports() == 0)
       {
	 printf("TalKT: No AX.25 port data avialable\n");       
	 return;
       }
   ax25init_done=1;

   strcpy(port, _port);

   if ((addr = ax25_config_get_addr(port)) == NULL)
   {
      printf("TalKT: Invalid port setting for port %s.\n", port);
      return;
   }

   ax25.fsa_ax25.sax25_family = AF_AX25;
   ax25.fsa_ax25.sax25_ndigis = 1;
   convert_call_entry(addr, ax25.fsa_digipeater[0].ax25_call);
   convert_call_entry(_call, ax25.fsa_ax25.sax25_call.ax25_call);
   addrlen = sizeof(struct full_sockaddr_ax25);
  
   if ((fd = socket(AF_AX25,SOCK_SEQPACKET,0)) < 0)
   {
      syslog(LOG_ERR,"socket: %s\n", strerror(errno));
      return;
   }
  
   if (bind(fd,(struct sockaddr *)&ax25,addrlen) < 0)
   {
      syslog(LOG_ERR,"bind: %s\n",strerror(errno));
      ::close(fd);
      return;
   }

   if (listen(fd, SOMAXCONN) < 0)
   {
      syslog(LOG_ERR,"listen: %s\n",strerror(errno));
      ::close(fd);
      return;
   }

   i = TRUE;
   ioctl(fd, FIONBIO, &i);



   strcpy(mycall, _call);
   Gross(mycall);
   sockrx = new QSocketNotifier( fd, QSocketNotifier::Read, this );
   connect( sockrx, SIGNAL(activated(int)), this, SLOT(rxFrame(int)) );
}


ListenAX25::~ListenAX25()
{
   ::close(fd);
   delete sockrx;
}


//   void Channel::rxFrame(int socket)
//
// Diese Funktion wird aufgerufen, wenn von unserer Verbindung Daten
// gekommen sind.
void ListenAX25::rxFrame(int socket)
{
#ifdef __GLIBC__
   socklen_t addrlen;
#else
   int addrlen;
#endif
   int fd;
   struct full_sockaddr_ax25 ax25;
   char call[100],*digis;
   char tmp[100],tmp2[100],tmp3[100];
   int k,i,len;
   char mycall[10];


   // Daten dieses Connects vom Kernel holen
   i = TRUE;
   ioctl(socket, FIONBIO, &i);

   addrlen = sizeof(ax25);
   if ((fd = accept(socket, (struct sockaddr *)&ax25, &addrlen)) == -1)
   {
      i = FALSE;
      ioctl(socket, FIONBIO, &i);
      return;
   }
   i = FALSE;
   ioctl(socket, FIONBIO, &i);

   i = TRUE;
   ioctl(fd, FIONBIO, &i);

   i = 0;

   // Rufzeichen der Gegenstation
   strcpy(call, ax25asc(&ax25.fsa_ax25.sax25_call));

   // SSID abhaengen und testen (!= 0)
   if ((i = POS('-',call)) != -1)
   {
      len = strlen(call)-i-1;
      COPY(tmp3,call,i+1,len);
      tmp3[len] = '\0';
      if (atoi(tmp3) == 0) call[i] = '\0';
   }


   // Digipeater
   if (ax25.fsa_ax25.sax25_ndigis > 0)
   {
      k = 0;
      tmp[0] = '\0';
      while (k < ax25.fsa_ax25.sax25_ndigis)
      {
         strcpy(tmp2, ax25asc(&ax25.fsa_digipeater[k]));
         


         // SSID abhngen und testen (!= 0)
         if ((i = POS('-',tmp2)) != -1)
         {
            len = strlen(tmp2)-i-1;
            COPY(tmp3,tmp2,i+1,len);
            tmp3[len] = '\0';
            if (atoi(tmp3) == 0) tmp2[i] = '\0';
         }

         if (k == 0)
            sprintf(tmp," via %s",tmp2);
         else
         {
            strcat(tmp," ");
            strcat(tmp,tmp2);
         }
         k++;
      }
      digis = (char *)strdup(tmp);
   }
   else
   {
      digis = (char *)malloc(1);
      digis[0] = '\0';
   }


   toplevel->connectionGot( port, mycall, call, digis, fd );
}


